// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_offensive_kick.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/common/logger.h>
#include <rcsc/player/player_predicate.h>

#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/bhv_shoot.h>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Bhv_BasicOffensiveKick::execute( PlayerAgent * agent )
{
	dlog.addText( Logger::TEAM,
			__FILE__": Bhv_BasicOffensiveKick" );

	const WorldModel & wm = agent->world();

	const PlayerPtrCont & opps = wm.opponentsFromSelf();
	const PlayerObject * nearest_opp
	= ( opps.empty()
			? static_cast< PlayerObject * >( 0 )
					: opps.front() );
	const double nearest_opp_dist = ( nearest_opp
			? nearest_opp->distFromSelf()
					: 1000.0 );
	const Vector2D nearest_opp_pos = ( nearest_opp
			? nearest_opp->pos()
					: Vector2D( -1000.0, 0.0 ) );

	const PlayerObject * nearest_opp_ahead=wm.opponentsFromSelf().front();



	PlayerPtrCont::const_iterator end=wm.opponentsFromSelf().end();
	for(PlayerPtrCont::const_iterator it=wm.opponentsFromSelf().begin();it!=end;it++)
	{
		if((*it)->pos().x>wm.self().pos().x+2)
			nearest_opp_ahead=(*it);
	}
	//wm.getOpponentsFromSelf();

	// = (const PlayerPtrCont &) (wm.getPlayerCont( (PlayerPredicate *) new XCoordinateForwardPlayerPredicate(wm.self().pos().x+2)));


	const double nearest_opp_ahead_dist = ( nearest_opp_ahead
				? nearest_opp_ahead->distFromSelf()
						: 1000.0 );
		const Vector2D nearest_opp_ahead_pos = ( nearest_opp_ahead
				? nearest_opp_ahead->pos()
						: Vector2D( -1000.0, 0.0 ) );

	double d=wm.self().pos().dist(ServerParam::i().theirTeamGoalPos());
	int prob=85;	//15%
	int dist=33;
	if( rand()%100>prob*(d-10)/(dist-10)) // prob for dist, 100% for 10
	{
		//	Vector2D disp(0,0);
		//	if(rand()%100>50)
			//		disp.y=-1*rand()%((int)round(ServerParam::i().DEFAULT_GOAL_WIDTH/2));
		//	else
		//		disp.y=rand()%((int)round(ServerParam::i().DEFAULT_GOAL_WIDTH/2)));
		dlog.addText( Logger::TEAM,
				__FILE__": (execute) shoot" );


		if (Bhv_Shoot2008().execute(agent))//ody_KickOneStep(ServerParam::i().theirTeamGoalPos()+disp,4,false))
		{

			return true;
		}
	}


	// dribble to my body dir
	if ( (nearest_opp_dist < 5.0
			&& nearest_opp_dist > ( ServerParam::i().tackleDist()
					+ ServerParam::i().defaultPlayerSpeedMax() * 1.5 )
					&& wm.self().body().abs() < 70.0)

	||
	(nearest_opp_ahead_dist > 5.0))
	{
		const Vector2D body_dir_drib_target
		= wm.self().pos()
		+ Vector2D::polar2vector(5.0, wm.self().body());

		int max_dir_count = 0;
		wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );

		if ( body_dir_drib_target.x < ServerParam::i().pitchHalfLength() - 1.0
				&& body_dir_drib_target.absY() < ServerParam::i().pitchHalfWidth() - 1.0
				&& max_dir_count < 3
		)
		{
			// check opponents
			// 10m, +-30 degree
			const Sector2D sector( wm.self().pos(),
					0.5, 10.0,
					wm.self().body() - 30.0,
					wm.self().body() + 30.0 );
			// opponent check with goalie
			if ( ! wm.existOpponentIn( sector, 10, true ) )
			{
				const int max_dash_step= wm.self().playerType().cyclesToReachDistance( wm.self().pos().dist( body_dir_drib_target ) );
				dlog.addText( Logger::TEAM,
						__FILE__": (execute) dribble to my body dir" );
				agent->debugClient().addMessage( "OffKickDrib(1)" );
				Body_Dribble( body_dir_drib_target,
						1.0,
						ServerParam::i().maxDashPower(),
						std::min( 5, max_dash_step )
				).execute( agent );
				agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
				return true;
			}
		}
	}



	Vector2D pass_point;
	int receiver=0;
	double first_speed=0.0;
	if (Body_Pass::get_best_pass( wm, &pass_point, &first_speed, &receiver) && wm.teammate(receiver))
	{
		Vector2D mate_goal=ServerParam::i().theirTeamGoalPos()-wm.teammate(receiver)->pos();
		int rand_range=rand()%7;	//between 0 and 6
		if(pass_point.x>0)
		{
			for(int range=2*rand_range;range>=0;range-=2)
			{
				if(! wm.existOpponentIn( Sector2D( pass_point,0.5, 0.5+range,mate_goal.dir() - 20.0,mate_goal.dir() + 20.0 ), 10, true ))
				{
					mate_goal.setLength(range/2);
					Vector2D recPnt=pass_point+mate_goal;
					Vector2D me_recPnt=recPnt-wm.self().pos();
					if(! wm.existOpponentIn( Sector2D( wm.self().pos(),0.5, me_recPnt.norm()+2,me_recPnt.dir() -10,me_recPnt.dir() + 10 ), 10, true ))
					{
						int kick_step = ( wm.gameMode().type() != GameMode::PlayOn
								&& wm.gameMode().type() != GameMode::GoalKick_? 1: 3 );
						if ( ! Body_SmartKick( recPnt,
								first_speed,
								first_speed * 0.96,
								kick_step ).execute( agent ) )
						{
							if ( wm.gameMode().type() != GameMode::PlayOn
									&& wm.gameMode().type() != GameMode::GoalKick_ )
							{
								first_speed = std::min( agent->world().self().kickRate() * ServerParam::i().maxPower(),
										first_speed );
								Body_KickOneStep(recPnt,
										first_speed
								).execute( agent );
								dlog.addText( Logger::ACTION,
										__FILE__": execute() one step kick" );
							}
							else
							{
								dlog.addText( Logger::ACTION,
										__FILE__": execute() failed to pass kick for range %d.",range );
								continue;
							}


						}

						if ( agent->config().useCommunication() && receiver != Unum_Unknown )
						{
							dlog.addText( Logger::ACTION,
									__FILE__": execute() set pass communication." );
							Vector2D target_buf = pass_point+mate_goal - wm.self().pos();
							target_buf.setLength( 1.0 );

							agent->addSayMessage( new PassMessage( receiver,
									recPnt + target_buf,
									agent->effector().queuedNextBallPos(),
									agent->effector().queuedNextBallVel() ) );
							return true;
						}
						break;
					}
				}
			}
		}

		if ( pass_point.x > wm.self().pos().x - 1.0 )
		{
			bool safety = true;
			const PlayerPtrCont::const_iterator opps_end = opps.end();
			for ( PlayerPtrCont::const_iterator it = opps.begin();
					it != opps_end;
					++it )
			{
				if ( (*it)->pos().dist( pass_point ) < 4.0 )
				{
					safety = false;
				}
			}

			if ( safety )
			{
				dlog.addText( Logger::TEAM,
						__FILE__": (execute) do best pass" );
				agent->debugClient().addMessage( "OffKickPass(1)" );
				Body_Pass().execute( agent );
				agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
				return true;
			}
		}
	}

	if ( nearest_opp_dist < 5.0 )
	{
		if ( Body_Pass().execute( agent ) )
		{
			dlog.addText( Logger::TEAM,
					__FILE__": (execute) do best pass" );
			agent->debugClient().addMessage( "OffKickPass(2)" );
			agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
			return true;
		}
	}


	Vector2D drib_target( 50.0, wm.self().pos().absY() );
	if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
	if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
	if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
	const AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();


	// opp is far from me
	if ( nearest_opp_dist > 5.0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": opp far. dribble(%.1f, %.1f)",
				drib_target.x, drib_target.y );
		agent->debugClient().addMessage( "OffKickDrib(4)" );
		Body_Dribble( drib_target,
				1.0,
				ServerParam::i().maxDashPower() * 0.8,
				5
		).execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	// opp is near

	// can pass
	if ( Body_Pass().execute( agent ) )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": (execute) pass",
				__LINE__ );
		agent->debugClient().addMessage( "OffKickPass(3)" );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	// opp is far from me
	if ( nearest_opp_dist > 3.0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": (execute) opp far. dribble(%f, %f)",
				drib_target.x, drib_target.y );
		const int max_dash_step
		= wm.self().playerType()
		.cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
		agent->debugClient().addMessage( "OffKickDrib(5)" );
		Body_Dribble( drib_target,
				1.0,
				ServerParam::i().maxDashPower() * 0.2,
				std::min( 5, max_dash_step )
		).execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}


	// opponent is behind me
	if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
	{
		// check opponents
		// 15m, +-30 degree
		const Sector2D sector( wm.self().pos(),
				0.5, 15.0,
				drib_angle - 30.0,
				drib_angle + 30.0 );
		// opponent check with goalie
		if ( ! wm.existOpponentIn( sector, 10, true ) )
		{
			const int max_dash_step
			= wm.self().playerType()
			.cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
			if ( wm.self().pos().x > 35.0 )
			{
				drib_target.y *= ( 10.0 / drib_target.absY() );
			}

			dlog.addText( Logger::TEAM,
					__FILE__": (execute) fast dribble to (%.1f, %.1f) max_step=%d",
					drib_target.x, drib_target.y,
					max_dash_step );
			agent->debugClient().addMessage( "OffKickDrib(2)" );
			Body_Dribble( drib_target,
					1.0,
					ServerParam::i().maxDashPower(),
					std::min( 5, max_dash_step )
			).execute( agent );
		}
		else
		{
			const int max_dash_step
			= wm.self().playerType()
			.cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
			dlog.addText( Logger::TEAM,
					__FILE__": (execute) slow dribble to (%.1f, %.1f)",
					drib_target.x, drib_target.y );
			agent->debugClient().addMessage( "OffKickDrib(3)" );
			Body_Dribble( drib_target,
					1.0,
					ServerParam::i().maxDashPower(),
					std::min( 5, max_dash_step )
			).execute( agent );

		}
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}


	if ( nearest_opp_dist > 2.5 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": hold" );
		agent->debugClient().addMessage( "OffKickHold" );
		Body_HoldBall().execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	{
		dlog.addText( Logger::TEAM,
				__FILE__": clear" );
		agent->debugClient().addMessage( "OffKickAdvance" );
		Body_AdvanceBall().execute( agent );
		agent->setNeckAction( new Neck_ScanField() );
	}

	return true;

}
