// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Majd Hawasly

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "domain_graph.h"
#include <cstdio>
#include <stdlib.h>
#include <sstream>


#include "sample_communication.h"
#include "strategy.h"
#include "rcsc/types.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/time/timer.h>
#include <rcsc/geom/polygon_2d.h>

#include <string>
#include <sstream>
#include <fstream>
#include <queue>
#include <utility>
#include <limits>
#include <cstdio>
#include <cmath>

#define DEBUG_PROFILE
// #define ACTION_CHAIN_DEBUG
// #define DEBUG_PAINT_EVALUATED_POINTS

//#define ACTION_CHAIN_LOAD_DEBUG
//#define DEBUG_COMPARE_SEARCH_TYPES

#ifdef DOMAIN_LOAD_DEBUG
#include <iostream>
#endif

using namespace rcsc;

//const size_t DomainGraph::DEFAULT_MAX_CHAIN_LENGTH = 4;
//const size_t DomainGraph::DEFAULT_MAX_EVALUATE_LIMIT = 500;

//std::vector< std::pair< Vector2D, double > > ActionChainGraph::S_evaluated_points;



/*-------------------------------------------------------------------*/
//const std::string DomainGraph::robustness_dir="./Robustness";


/*!

 */
DomainGraph::DomainGraph( )
{
    //instance();	//initialize the static member
    role_num=0;
    rScore=0;
    lScore=0;
    M_action= boost::shared_ptr<const CooperativeAction>();
    chosen_link=CooperativeAction::NoAction;
    unlearnt_link=CooperativeAction::NoAction;
    loaded_Robustness=false;
    best_link=-1;


    for (int i=0;i<NUM_TEAM_STATES;i++)
    {
        V[i]=0.5;
    }
    V[0]=0;	//Unknown
    V[13]=0;	//Lost

    if (!readRobustness(role_num))
    {
        for (int i=0;i<NUM_TEAM_STATES;i++)
        {
            for (int j=0;j<NUM_LOCAL_GAMES;j++)
            {
                for (int h=0;h<NUM_TEAM_STATES;h++)
                {
                    for (int f=0;f<NUM_FEATURES;f++)
                    {
                        if (h==13)	//Lost
                            R[i][j][h][f]=0.5;
                        else
                            R[i][j][h][f]=0.5/(NUM_LOCAL_GAMES-1);
                    }
                    //	if(j==0 & i==h)	//hold_ball
                    //	    R[i][j][h]=10;
                }
            }
        }
    }

    for (int t=0;t<NUM_TEAM_STATES;t++)
    {
        lossTstates[t]=false;
        goalTstates[t]=false;
    }

    for (uint g=0;g<NUM_LOCAL_GAMES;g++)
    {
        features[g]=-1;
        f_times[g]=-1;
    }

    lossTstates[13]=true;	//Tstate 13: Loss

//     goalTstates[0]=false;	//Tstate 0: Unknown
//     goalTstates[13]=false;	//Tstate 13: Loss
//Later on -- only Goal state would be true and all other will be false
    goalTstates[12]=true;

}



void DomainGraph::loadRobustness()
{
 if (!loaded_Robustness)
    {
        if (!readRobustness(role_num))
        {
            for (uint i=0;i<NUM_TEAM_STATES;i++)
            {
                for (uint j=0;j<NUM_LOCAL_GAMES;j++)
                {
                    for (uint h=0;h<NUM_TEAM_STATES;h++)
                    {
                        for (uint f=0;f<NUM_FEATURES;f++)
                        {
                            if (h==13)	//Lost
                                R[i][j][h][f]=0.5;
                            else
                                R[i][j][h][f]=0.5/(NUM_LOCAL_GAMES-1);
                        }
                        //	if(j==0 & i==h)	//hold_ball
                        //	    R[i][j][h]=10;
                    }
                }
            }
        }
        loaded_Robustness=true;
    }
    
}

/*-------------------------------------------------------------------*/

bool DomainGraph::readRobustness(int role)
{
    std::string rr;
    std::stringstream r;
    r << role;
    rr = r.str();
//     char r[2];
//     itoa(role,r,10);
    //check the role number to get the right weights
    static bool r_initialized = false;

    if ( r_initialized )
    {
        std::cerr << __FILE__ << ' ' << __LINE__ << ": already initialized."
                  << std::endl;
        return false;
    }

    std::string configpath = "Robustness";
    if ( ! configpath.empty() && configpath[ configpath.length() - 1 ] != '/' )
    {
        configpath += '/';
    }


    std::string filepath=configpath + "R"+rr;

    std::ifstream fin( filepath.c_str() );
    if ( ! fin.is_open() )
    {
        std::cerr << __FILE__ << ':' << __LINE__ << ':'
                  << " ***ERROR*** failed to open file [" << filepath << "]"
                  << std::endl;
        return false;
    }

    fin.seekg( 0 );

    for (int i=0;i<NUM_TEAM_STATES;i++)
        for (int g=0;g<NUM_LOCAL_GAMES;g++)
            for (int j=0;j<NUM_TEAM_STATES;j++)
                for (int f=0;f<NUM_FEATURES;f++)
                {
                    //
                    // read data from file
                    //
                    char line[10];
                    fin.getline(line,9);
                    double r=std::atof(line);
                    R[i][g][j][f]=r;
                }
// 	if ( !  )
// 	{
// 	    std::cerr << __FILE__ << ':' << __LINE__ << ':'
// 	    << " ***ERROR*** failed to read formation [" << filepath << "]"
// 									<< std::endl;
// 									f.reset();
// 									return f;
// 	}
//

// 	if (  )
// 	{
// 	    std::cerr << "Failed to read Robustness file" << std::endl;
// 	    return false;
// 	}

    r_initialized = true;
    return true;
}


/*-------------------------------------------------------------------*/


bool DomainGraph::writeRobustness(int role)
{
    std::string rr;
    std::stringstream r;
    r << role;
    rr = r.str();


    std::string configpath = "Robustness";
    if ( ! configpath.empty() && configpath[ configpath.length() - 1 ] != '/' )
    {
        configpath += '/';
    }


    std::string filepath=configpath + "R"+rr;	//add role number here

    std::ofstream fout( filepath.c_str() );
    if ( ! fout.is_open() )
    {
        std::cerr << __FILE__ << ':' << __LINE__ << ':'
                  << " ***ERROR*** failed to open file [" << filepath << "]"
                  << std::endl;
        std::cout<<"----------Failed to open file-------------"<<std::endl;
        return false;
    }

    std::cout<<"----------Writing to file "<< filepath<<"-------------"<<std::endl;
    fout.seekp(0);

    for (uint i=0;i<NUM_TEAM_STATES;i++)
        for (uint g=0;g<NUM_LOCAL_GAMES;g++)
            for (uint j=0;j<NUM_TEAM_STATES;j++)
                for (uint f=0;f<NUM_FEATURES;f++)
                {
                    //
                    // write data to file
                    fout<<R[i][g][j][f]<<std::endl;
                }

    return true;
}
/*-------------------------------------------------------------------*/

void DomainGraph::findBestLink(const rcsc::WorldModel& wm, int toPassTo)
{
    dlog.addText ( Logger::TEAM,
                   __FILE__": starting DG search" );
    //use the domain graph and the current world state to choose which game to choose
    double max_v=0;
    int blink=-1;
    // for (int i=0;i<NUM_TEAM_STATES;i++)		//computed value for owned states -- new loop so all updates can be incorporated

    if (/*i==M_current_Tstate  &&*/  ownedTstates[M_current_Tstate])//active(i))
    {
        int i=M_current_Tstate;
        //go through all the links to find the best value that can be attained from Tstate i
        if (i==13)	//Lost games
            gameOutcome(i,(int)CooperativeAction::GetBall,wm, toPassTo,max_v, blink);	//only getBall
        else
        {
            for (int j=0;j<NUM_LOCAL_GAMES;j++)
            {
		if(j==(int)CooperativeAction::GetBall) continue;
                gameOutcome(i,j,wm, toPassTo,max_v, blink);
            }
        }
    }
//                 int f=extractFeatures(j,wm,toPassTo);	//extract the right features of the game j
//                 double max_r=0;
//                 int outcome=-1;
//                 double sum_ij=0;
//                 for (int h=0;h<NUM_TEAM_STATES;h++)
//                 {
//                     //find the most reliable outcome of game j
//                     if (R[i][j][h][f]>ROBUSTNESS_THRESHOLD && R[i][j][h][f]>max_r)
//                     {
//                         max_r=R[i][j][h][f];
//                         outcome=h;
//                     }
//                     sum_ij+=R[i][j][h][f];
//                 }
//                 //game j leads most likely to state h with robustnesss max_r
//                 if (outcome!=-1)
//                 {
//                     if (V[outcome]*max_r/sum_ij>max_v)
//                     {
//                         max_v=V[outcome]*max_r/sum_ij;
//                         blink=j;
//                     }
//                 }


    best_link = blink;
   /* if (blink==-1)
        blink=rand()%5 ;*/	//pick a random action for now


    if (wm.self().isKickable() || ownedTstates[M_current_Tstate])	//Only if I have the control over the ball, I should decide what to do next
    {
	if (blink==-1);// blink=rand()%4 ;	//pick a random action for now {0,1,2,3}
        switch (blink)
        {
        case 0://H
            chosen_link=CooperativeAction::Hold;
            break;

        case 1://D
            chosen_link=CooperativeAction::Dribble;
            break;

        case 2://P
            chosen_link=CooperativeAction::Pass;
            break;

        case 3://SH
            chosen_link=CooperativeAction::Shoot;
            break;

        default:
            chosen_link=CooperativeAction::NoAction;
        }
    }
    else //if (blink==4)//GB
    {
        chosen_link=CooperativeAction::GetBall;
	blink=4;
    }


    best_link=blink;

    //std::cout<<"Best game is: "<< best_link<<std::endl;

    dlog.addText ( Logger::TEAM,
                   __FILE__": DG search finished!" );


}


/*-------------------------------------------------------------------*/

//starting state i, game j --> outcome
void DomainGraph::gameOutcome(int s,int g,const rcsc::WorldModel& wm, int toPassTo,double& max_v, int& blink)
{
    int f=extractFeatures(g,wm,toPassTo);	//extract the right features of the game j
    double max_r=0;
    int outcome=-1;
    double sum_sg=0;
     if (f!=-1)
    {
    for (int h=0;h<NUM_TEAM_STATES;h++)
    {
        //find the most reliable outcome of game g
        if (R[s][g][h][f]>ROBUSTNESS_THRESHOLD && R[s][g][h][f]>max_r)
        {
            max_r=R[s][g][h][f];
            outcome=h;
        }
        //sum_sg+=R[s][g][h][f];
    }
    }
    //game g leads most likely to state outcome with robustnesss max_r
    if (outcome!=-1)
    {
	dlog.addText ( Logger::TEAM,
		       __FILE__": Robustness of game %d from %d is %02f",g,s,V[outcome]*max_r );
        if (V[outcome]*max_r>max_v) // /sum_sg
        {
            max_v=V[outcome]*max_r;// /sum_sg;
            blink=g;
        }
    }
}

/*-------------------------------------------------------------------*/

void DomainGraph::setAction(CooperativeAction::ConstPtr action)
{
    //std::cout<<"setting action..."<<std::endl;
    if (action)// && sizeof(*action)==sizeof(CooperativeAction))	//make sure that the pointer is pointing to an action
    {
        M_action=action;
    }
    else
        M_action=boost::shared_ptr<const CooperativeAction>();
}

/*-------------------------------------------------------------------*/

/*!

 */
DomainGraph &
DomainGraph::instance()
{
    static DomainGraph dg_instance;
    return dg_instance;
}

/*-------------------------------------------------------------------*/


// /*
// Hold, 0
// Dribble, 1
// Pass, 2
// Shoot, 3
// Clear, 4
// Move, 5
// 	//6
// NoAction,//7
// getball	//8 or 4 ?
// */
void
DomainGraph::update(const rcsc::WorldModel & wm, int (&heardVs)[12])
{

    dlog.addText ( Logger::TEAM,
                   __FILE__": starting DG update" );

    if (!ownedTstates[wm.self().unum()])	//update this when you know unum
    {
        for (uint i=0;i<NUM_TEAM_STATES;i++)
            ownedTstates[i]=false;
        ownedTstates[wm.self().unum()]=true;
        ownedTstates[13]=true;		//state Lost -- everyone should be active
    }

    //save the old team state before updating
    M_old_Tstate=M_current_Tstate;

    //update using world model
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    int our_min = std::min( self_min, mate_min );


    if ((wm.ourSide()==rcsc::LEFT && wm.gameMode().scoreLeft()>lScore) ||  (wm.ourSide()==rcsc::RIGHT && wm.gameMode().scoreRight()>rScore))
    {
        //ourScore+=1;
        M_current_Tstate = 12;	//Goal
        dlog.addText ( Logger::TEAM,
                       __FILE__": TState: GOAL!" );
    }

    else
    {
    	//opponent is closer and can kick
        if ( opp_min <= our_min - 2  && wm.existKickableOpponent())//&& opp_min<POSS_THRESHOLD   //only if opponents can kick the ball it becomes L
        {
            dlog.addText ( Logger::TEAM,
                           __FILE__": TState: Lost" );
            M_current_Tstate = 13;	//Lost
        }

    	//opponent is not closer or cannot kick
        else
        {
        	//we are clearly closer
            if ( our_min <= opp_min - 2  ||  (our_min <= opp_min &&  M_old_Tstate!=0 &&  M_old_Tstate!=13))
            {
            	//I'm closer
                if ( self_min <= mate_min - 2 && (self_min<POSS_THRESHOLD  || M_old_Tstate == wm.self().unum()))
                {
                    dlog.addText ( Logger::TEAM,
                                   __FILE__": Tstate: Hold" );
                    M_current_Tstate = wm.self().unum();	//H
                }
                //mate is closer
                else if ( mate_min <= self_min - 2 && (mate_min<POSS_THRESHOLD || M_old_Tstate == wm.interceptTable()->fastestTeammate()->unum()))
                {

                    M_current_Tstate = wm.interceptTable()->fastestTeammate()->unum();	//H
                    dlog.addText ( Logger::TEAM,
                                   __FILE__": Tstate: Hold%d",M_current_Tstate );
                }
            }
            //nobody is that close
            else
            {
            	dlog.addText ( Logger::TEAM,
            			__FILE__": Tstate: unknown -- should not have changed: %d",M_old_Tstate);
            	M_current_Tstate = M_old_Tstate;	//unknown   ////////////////////////////////////////////
            }
        //}
            //we are not directly closer, neither the opponent
//            else
//            {
//                dlog.addText ( Logger::TEAM,
//                               __FILE__": Tstate: unknown" );
//                M_current_Tstate = 0;	//unknown
//            }
        }
    }

    //std::cout<<"current_TState is: "<<M_current_Tstate<<std::endl;


    learnRobustness(wm);
    updateValues(wm, heardVs);

    //seeing if not reseting this is better

    //chosen_link=CooperativeAction::NoAction;

    if (wm.gameMode().scoreLeft()!=lScore || wm.gameMode().scoreRight()!=rScore)
    {
    	 lScore=wm.gameMode().scoreLeft();
    	        rScore=wm.gameMode().scoreRight();
    	        //////////////////////////////////////////////////////
    	        writeRobustness(role_num);	//consider specifiying a file per-role or unum if roles are static
    	    }   //////////////////////////////////////////////////////

    	    dlog.addText ( Logger::TEAM,
    	                   __FILE__": DG update finished!" );
}


void DomainGraph::updateValues(const rcsc::WorldModel& wm, int (&heardValues)[12])
{
    dlog.addText ( Logger::TEAM,
                   __FILE__": starting V update" );

    for (uint i=0;i<NUM_TEAM_STATES;i++)
    {
        if (lossTstates[i])	//fixed value for loss states
            V[i]=0;
        else if (goalTstates[i])	//fixed value for goal states
            V[i]=1;	//generating more value in the system

        if (i>=1 && i<=11)
        {
            if (heardValues[i]!=-1)//heardTstates[i]=-1)//heardTstates[i]!=0)		//value communicated by a mate
                V[i]=std::min(1.0,heardValues[i]/100.0+0.5);//heardTstates[i];
            else	// no message from the player
                V[i]=(0.5*V[i]<0.5?0.5*V[i]:(rand()%100)/200.0);	//discount
        }
    }

    for (uint i=0;i<NUM_TEAM_STATES;i++)		//compute value for owned states -- new loop so all received updates can be incorporated
    {
        if (ownedTstates[i] && i!=13)
        {
            double max_v=0;

            //go through all the links to find the best value that can be attained from Tstate i
            for (uint j=0;j<NUM_LOCAL_GAMES;j++)
            {
                if (j!=2)	//not a Pass
                {
                    int f=extractFeatures(j,wm);
		    dlog.addText ( Logger::TEAM,
				   __FILE__": features -- %d = %d",j,f);
		    

                    //double sum_ij=0;
//                 for (int h=0;h<NUM_TEAM_STATES;h++)
//                 {
//                     sum_ij+=R[i][j][h][f];
//                 }

                    if (f!=-1)
                    {
                        for (uint h=0;h<NUM_TEAM_STATES;h++)
                        {
                            if (R[i][j][h][f]>ROBUSTNESS_THRESHOLD)//sum_ij!=0 && /sum_ij
                            {
                                if (R[i][j][h][f]*V[h]>max_v)// /sum_ij
                                {
                                    max_v=R[i][j][h][f]*V[h]; // /sum_ij
                                    //best_link=j;
                                }
                            }
                        }
                    }
                }
                else //a Pass
                {
                    for (uint h=0;h<NUM_TEAM_STATES;h++)
                    {
                        if (h==i || h<1 || h>11) continue;
                        int f=extractFeatures(j,wm,h);
			
			dlog.addText ( Logger::TEAM,
				       __FILE__": features -- Passing to %d = %d",h,f);
			
                        if (f!=-1 && R[i][j][h][f]>ROBUSTNESS_THRESHOLD)
                        {
                            if (R[i][j][h][f]*V[h]>max_v)
                            {
                                max_v=R[i][j][h][f]*V[h];
                            }
                        }

                    }
                }
            }
            V[i]=max_v;
            if(V[i]>0.5)valueMsg=new ValueMessage(i,V[i]);
            else valueMsg=0;
        }
    }

    //send computed values using communication to neighbour agents
    dlog.addText ( Logger::TEAM,
                   __FILE__": V update finished!" );
    dlog.addText ( Logger::TEAM,
		   __FILE__": Values: %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f",
		   V[0],V[1],V[2],V[3],V[4],V[5],V[6],V[7],V[8],V[9],V[10],V[11],V[12],V[13] );

}
//}
//			}
//
//		    }
//		}
//            }
//            V[i]=max_v;
//            valueMsg=new ValueMessage(i,V[i]);
//        }
//    }
//
//
//    //send computed values using communication to neighbour agents
//    dlog.addText ( Logger::TEAM,
//                   __FILE__": V update finished!" );
//    dlog.addText ( Logger::TEAM,
//		   __FILE__": Values: %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f %02f",
//		   V[0],V[1],V[2],V[3],V[4],V[5],V[6],V[7],V[8],V[9],V[10],V[11],V[12],V[13] );

//}

/*-------------------------------------------------------------------*/

int DomainGraph::extractFeatures(int g, const rcsc::WorldModel & wm)//const rcsc::WorldModel & wm)
{
    if (f_times[g]==wm.time().cycle())
        return features[g];

    if (g==0 || g==1)	//Hold or Dribble
        return feature_hd(wm);
   // else if (g==2)	//Pass
   //     return feature_p(wm);
    else if (g==3)	//Shoot
        return feature_sh(wm);
    else if (g==4)	//getball
        return feature_gb(wm);
    else
        return -1;
}

/*-------------------------------------------------------------------*/

int DomainGraph::extractFeatures(int g, const rcsc::WorldModel & wm, int teammate_unum)//const rcsc::WorldModel & wm)
{
    if (g==2 && teammate_unum!=0)	//Pass
    {
        return feature_p(wm, teammate_unum);
    }
    else return extractFeatures(g, wm);
}

/*-------------------------------------------------------------------*/

int DomainGraph::feature_hd(const rcsc::WorldModel & wm)
{
    int c=wm.countOpponentsIn<Circle2D>(Circle2D(wm.self().pos(),20),100,true);
    if (c>2)c=2;

    features[0]=c;	//set it for reuse
    f_times[0]=wm.time().cycle();
    features[1]=c;
    f_times[1]=wm.time().cycle();
    return c;
}

/*-------------------------------------------------------------------*/

int DomainGraph::feature_gb(const rcsc::WorldModel & wm)
{
    int c=wm.countOpponentsIn<Circle2D>(Circle2D(wm.ball().pos(),20),100,true);
    if (c>2)c=2;

    features[4]=c;	//set it for reuse
    f_times[4]=wm.time().cycle();
    return c;
}

/*-------------------------------------------------------------------*/
int DomainGraph::feature_p(const rcsc::WorldModel & wm)
{
    for (int i=1;i<=11;i++)
    {
        if (i==wm.self().unum())
            continue;
    //feature_p(wm,i);
    }
}
int DomainGraph::feature_p(const rcsc::WorldModel & wm, int teammate_unum)
{
    if(!wm.teammate(teammate_unum)|| wm.teammate(teammate_unum)->distFromSelf()>MAX_BALL_RANGE)
	return -1;//////////////
    //Vector2D p1=wm.self().pos();
    double p1x=wm.self().pos().x;
    double p1y=wm.self().pos().y;
    //p1.reverse();
    //Vector2D p2=wm.teammate(teammate_unum)->pos();
    double p2x=wm.teammate(teammate_unum)->pos().x;
    double p2y=wm.teammate(teammate_unum)->pos().y;
    
    double p3x=p2x-p1x;
    double p3y=p2y-p1y;
    //p1.add(p2);	//p2-p1
    //Vector2D p3=p2-p1;
    double p31x=-1*p3y;
    double p31y=p3x;
    
    p31x/=sqrt(p31x*p31x+p31y*p31y)/5;
    p31y/=sqrt(p31x*p31x+p31y*p31y)/5;
    
    
    double p4x=p2x+p31x;
    double p4y=p2y+p31y;
    
    double p32x=p3y;
    double p32y=-1*p3x;
    
    p32x/=sqrt(p32x*p32x+p32y*p32y)/5;
    p32y/=sqrt(p32x*p32x+p32y*p32y)/5;
    
    double p5x=p2x+p32x;
    double p5y=p2y+p32y;
    
    //Vector2D p31=Vector2D(-1*p1.y,p1.x);
    //p31.normalize();
    ///!!!!scaling needed here
   // Vector2D p4=p2;
   // p4.add(p31);	//p2+p31
//    Vector2D p32=Vector2D(-1*p1.y,p1.x);
  //  p32.normalize();
    ///!!!!scaling needed here
  //  Vector2D p5=p2;
 //   p5.add(p32);	//p2+p32
//
    int c=wm.countOpponentsIn<Triangle2D>(Triangle2D(Vector2D(p1x,p1y),Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);
    if (c>2)c=2;

    features[2]=c;
    f_times[2]=wm.time().cycle();
    return c;
}

int DomainGraph::feature_sh(const rcsc::WorldModel & wm)
{
    if(wm.self().pos().dist(ServerParam::instance().theirTeamGoalPos()) > MAX_BALL_RANGE)
	return -1;
    
    double p1x=wm.self().pos().x;
    double p1y=wm.self().pos().y;
    //p1.reverse();
    //Vector2D p2=wm.teammate(teammate_unum)->pos();
    double p2x=ServerParam::i().theirTeamGoalPos().x;
    double p2y=ServerParam::i().theirTeamGoalPos().y;
    
    double p3x=p2x-p1x;
    double p3y=p2y-p1y;
    //p1.add(p2);	//p2-p1
    //Vector2D p3=p2-p1;
    double p31x=-1*p3y;
    double p31y=p3x;
    
    p31x/=sqrt(p31x*p31x+p31y*p31y)/ServerParam::i().goalWidth()/2;
    p31y/=sqrt(p31x*p31x+p31y*p31y)/ServerParam::i().goalWidth()/2;
    
    
    double p4x=p2x+p31x;
    double p4y=p2y+p31y;
    
    double p32x=p3y;
    double p32y=-1*p3x;
    
    p32x/=sqrt(p32x*p32x+p32y*p32y)/ServerParam::i().goalWidth()/2;
    p32y/=sqrt(p32x*p32x+p32y*p32y)/ServerParam::i().goalWidth()/2;
    
    double p5x=p2x+p32x;
    double p5y=p2y+p32y;
    
    
//     //ServerParam::i().goalWidth()/2)
//     Vector2D p1=wm.self().pos();
//     p1.reverse();
//     Vector2D p2=ServerParam::i().theirTeamGoalPos();
//     p1.add(p2);	//p2-p1
//     //Vector2D p3=p2-p1;
//     Vector2D p31=Vector2D(-1*p1.y,p1.x);
//     p31.normalize();
//     p31.scale(;
//     Vector2D p4=p2;
//     p4.add(p31);	//p2+p31
//     p4.scale(ServerParam::i().goalWidth()/2);
//     Vector2D p32=Vector2D(-1*p1.y,p1.x);
//     p32.normalize();
//     p32.scale(ServerParam::i().goalWidth()/2);
//     Vector2D p5=p2;
//     p5.add(p32);	//p2+p32


    int c=wm.countOpponentsIn<Triangle2D>(Triangle2D(Vector2D(p1x,p1y),Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);
    if (c>2)c=2;

    features[3]=c;
    f_times[3]=wm.time().cycle();
    return c;

}

/*-------------------------------------------------------------------*/

void DomainGraph::learnRobustness(const rcsc::WorldModel & wm)
{
    //std::cout<<"Unlearnt link: "<<unlearnt_link<<std::endl;
    if (unlearnt_link!=CooperativeAction::NoAction && M_current_Tstate!=unlearnt_link_src && M_current_Tstate!=0) // not the same and not unknown
    {
        dlog.addText ( Logger::TEAM,
                       __FILE__": learning unlearnt_link...%d",(int)unlearnt_link );
        learnRobustness(unlearnt_link,unlearnt_link_pars,wm,unlearnt_link_src);	//learn the delayed action
        unlearnt_link=CooperativeAction::NoAction;	//clear the unlearnt action
        dlog.addText ( Logger::TEAM,
                       __FILE__": unlearnt_link cleared!" );

    }



    if (active(wm.self().unum())
            && (chosen_link==CooperativeAction::Pass || chosen_link==CooperativeAction::Shoot)
            && (M_current_Tstate==M_old_Tstate || M_current_Tstate==0)
            && M_old_Tstate!=0) // || chosen_link==CooperativeAction::GetBall )
    {
        unlearnt_link=chosen_link;	//keep this until the outcome is known
        unlearnt_link_src=M_old_Tstate;
        unlearnt_link_pars=(chosen_link==CooperativeAction::Pass && M_action && wm.teammate(M_action->targetPlayerUnum()))?M_action->targetPlayerUnum():0;
        dlog.addText ( Logger::TEAM,
                       __FILE__": setting unlearnt-link: %d",(int)unlearnt_link);
    }
    else
    {

        if (chosen_link==CooperativeAction::Pass && M_action && wm.teammate(M_action->targetPlayerUnum()))
            learnRobustness(chosen_link,M_action->targetPlayerUnum(),wm);
        else
            learnRobustness(chosen_link,0,wm);
    }
}

//     //use M_current_Tstate and M_old_Tstate with addition to chosen_link to update R matrix
//     //use also the world model to extract features suitable to parametrise the actions
//     dlog.addText ( Logger::TEAM,
//                    __FILE__": starting R update" );
//
//     int f=-1;
//     if (chosen_link!=CooperativeAction::NoAction)	//&& M_Action
//     {
//         if (chosen_link==CooperativeAction::Hold || chosen_link==CooperativeAction::Dribble )
//         {
//             f=extractFeatures((int)chosen_link,wm);
//         }
//         else
//         {
//             if (chosen_link==CooperativeAction::Pass && M_action && wm.teammate(M_action->targetPlayerUnum()))//M_action->targetPlayerUnum()>0 &&  M_action->targetPlayerUnum()!=M_action->playerUnum())
//             {
//                 f=extractFeatures((int)chosen_link,wm,M_action->targetPlayerUnum());
//             }
//             else
//             {
//                 if (chosen_link==CooperativeAction::Shoot)
//                 {
//                     f=extractFeatures((int)chosen_link,wm);
//                 }
//             }
//         }
//
//         //actual update
//         R[M_old_Tstate][(int)chosen_link][M_current_Tstate][f]+=1;
//
//
// 	std::cout<<"Learning: "<<M_old_Tstate<<" -- "<<(int)chosen_link<<" -> "<<M_current_Tstate<<std::endl;
//
// 	if(M_current_Tstate==12)
// 	chosen_link=CooperativeAction::NoAction;
//     }
//
//     dlog.addText ( Logger::TEAM,
//                    __FILE__": DG update finished!" );

/*-------------------------------------------------------------------*/

void DomainGraph::learnRobustness(CooperativeAction::ActionCategory alink, int mate_unum, const rcsc::WorldModel & wm)
{
    if (!active(wm.self().unum()))
        return;
    learnRobustness(alink, mate_unum, wm, M_old_Tstate);
}

void DomainGraph::learnRobustness(CooperativeAction::ActionCategory alink, int mate_unum, const rcsc::WorldModel & wm, int src_Tstate)
{    //use M_current_Tstate and M_old_Tstate with addition to chosen_link to update R matrix
    //use also the world model to extract features suitable to parametrise the actions

    dlog.addText ( Logger::TEAM,
		   __FILE__": starting R update" );

    int f=-1;
    if (alink!=CooperativeAction::NoAction)
    {
        if (alink==CooperativeAction::Hold || alink==CooperativeAction::Dribble )
        {
            f=extractFeatures((int)alink,wm);
        }
        else if (alink==CooperativeAction::Pass && mate_unum!=0)
        {
            f=extractFeatures((int)alink,wm,mate_unum);
        }
        else if (alink==CooperativeAction::Shoot)
        {
            f=extractFeatures((int)alink,wm);
        }
        else if (alink==CooperativeAction::GetBall)	//GetBall
        {
            f=extractFeatures((int)alink,wm);
        }

        //actual update
        //if ((int)link==7)R[M_old_Tstate][4][M_current_Tstate][f]+=1;
        //else
        if (f!=-1)
        {
        R[src_Tstate][(int)alink][M_current_Tstate][f]+=ALPHA;

        double sum_d=0;
            for (uint d=0;d<NUM_TEAM_STATES;d++)
        {
            sum_d+=R[src_Tstate][(int)alink][d][f];
        }
        if (sum_d!=0)
        {
                for (uint d=0;d<NUM_TEAM_STATES;d++)
            {
                R[src_Tstate][(int)alink][d][f]/=sum_d;
            }
        }


        dlog.addText ( Logger::TEAM,
                       __FILE__": Learning: %d --%d-> %d", src_Tstate,(int)alink,M_current_Tstate);
	//std::cout<<"Learning: "<<src_Tstate<<" --"<<(int)alink<<"-> "<<M_current_Tstate<< std::endl;
        //std::cout<<"Learning: "<<M_old_Tstate<<" -- "<<(int)chosen_link<<" -> "<<M_current_Tstate<<std::endl;
        }

        if (M_current_Tstate==12)
            chosen_link=CooperativeAction::NoAction;


    }

    dlog.addText ( Logger::TEAM,
                   __FILE__": DG update finished!" );
}



/*-------------------------------------------------------------------*/


PredicateComplex::PredicateComplex()

{
    //o_activePredicates.clear();
    //d_activePredicates.clear();
    //o_activePredicates.push_back(open);
    M_action= boost::shared_ptr<const CooperativeAction>();
    pass_mate=-1;
    marked_unum=-1;
    blocking=false;
    markingMsg=new MarkingMessage(-1);

    for (uint i=0;i<=11;i++)
    {
    	M[i]=-1;
    	MT[i]=-1;
    }
}


PredicateComplex &
PredicateComplex::instance()
{
    static PredicateComplex pc_instance;
    return pc_instance;
}



//this functions generates and optimises between possible targets that achieve the active predicates
void PredicateComplex::generateTargets(const rcsc::WorldModel& wm,std::list<rcsc::Vector2D>& targets,bool offence)
{
	if((offence && o_activePredicates.size()==0) || (!offence && d_activePredicates.size()==0 ))
		return;

	dlog.addText ( Logger::TEAM,__FILE__": generating targets...");

	Vector2D v1, v2;
	v1.setPolar(5,wm.self().body()+AngleDeg(60));
	v2.setPolar(5,wm.self().body()+AngleDeg(-60));

	int dd=12;
	for(;dd>=6;dd=dd-2)
	{
		v1.setLength(dd);
		v2.setLength(dd);
		if(!wm.existOpponentIn(Sector2D(wm.self().pos(),0.5,dd,wm.self().body()-45,wm.self().body()+45),100,true))
			break;
	}

	int MAX_REACHABLE_DIST=dd/2;

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	int our_min = std::min( self_min, mate_min );


	//generate a random set of reachable points
	//std::list<rcsc::Vector2D> targets;

	for(int j=0; j<100; j++)
	{
		//generate a reachable random target
		int dr=2;
		if(MAX_REACHABLE_DIST>2)dr+=rand()%(MAX_REACHABLE_DIST-2);
		int dth=(((rand()%2)*2)-1) * rand()%(45);

		if(dr<2) continue;

		double dx=dr*cos((wm.self().body().degree() +dth)/180*3.14);
		double dy=dr*sin((wm.self().body().degree() +dth)/180*3.14);



		//push it into vector targets
		rcsc::Vector2D vec=Vector2D(dx+wm.self().pos().x,dy+wm.self().pos().y);
		rcsc::Vector2D vec2=Vector2D(dx,dy);

		//in-field?

		if(abs(vec.x) < (ServerParam::instance().DEFAULT_PITCH_LENGTH/2 -1) && abs(vec.y)< (ServerParam::instance().DEFAULT_PITCH_WIDTH/2-1))
		{
			targets.push_back(vec);
			//if(wm.self().unum()==3) std::cout<<"New point ("<<vec.x<<", "<<vec.y<<")"<<std::endl;
		}
	}

	if(o_activePredicates.size()==0) return;

	int d[o_activePredicates.size()];

	for(uint j=0;j<o_activePredicates.size();j++)
		d[j]=0;



	if(offence)	//we have the ball
	{
		for(uint i=0;i<o_activePredicates.size();i++)
		{
			switch(o_activePredicates[i])
			{
			case open:
			{
				std::list<rcsc::Vector2D>::iterator j=targets.begin();
				//int c=targets.size();
				//std::cout<<"Open: "<<c<<std::endl;
				while(true)
				{

					if(j==targets.end()) break;
					//check locations that keep you open
					if(!isOpen(*j,wm))
					{
						//std::list<rcsc::Vector2D>::iterator temp=j;

						targets.erase(j);
						d[i]++;
						//std::cout<<--c<<std::endl;
						//c--;
						j--;
						//j=temp;
					}
					j++;
				}
				break;
			}
			case contact:
			{
				std::list<rcsc::Vector2D>::iterator j=targets.begin();
				//int c=targets.size();
				//std::cout<<"Contact: "<<c<<std::endl;
				while(true)
				{
					if(j==targets.end()) break;
					//check locations that keep contact
					if(!isContact(*j,wm))
					{
						//std::list<rcsc::Vector2D>::iterator temp=j;

						targets.erase(j);
						d[i]++;
						//std::cout<<--c<<std::endl;
						//c--;
						j--;
						//j=temp;
					}
					j++;

				}
				break;
			}
			case position:
			{
				//				std::vector<rcsc::Vector2D> v_targets(targets.size());

				//				copy(targets.begin(), targets.end(), v_targets.begin());
				std::list<rcsc::Vector2D>::iterator j=targets.begin();
				//int c=targets.size();
				int e=0;
				//std::cout<<"Position: "<<c<<std::endl;
				while(true)
				{
					//e++;
					if(j==targets.end()) break;
					//check locations that keep contact
					if(!isPosition(*j,wm))
					{
						//	dlog.addText ( Logger::TEAM,__FILE__": targets not good -- %d",i);
						//						std::cout<<"targets j-1 "<<(*(j--)).x<<","<<(*(j)).y
						//								<<"-- j "<<(*(j++)).x<<","<<(*(j)).y
						//								<<"-- j+1 "<<(*(j++)).x<<","<<(*(j)).y<<std::endl;
						//						j--;
						//						dlog.addText ( Logger::TEAM,__FILE__": targets %d (%f,%f) -- %d (%f,%f) -- %d (%f,%f)",
						//								i-1,v_targets[i-1].x,v_targets[i-1].y,
						//								i,v_targets[i].x,v_targets[i].y,
						//								i+1,v_targets[i+1].x,v_targets[i+1].y);
						//					std::list<rcsc::Vector2D>::iterator temp=j;

						//j--;
						targets.erase(j);
						d[i]++;

						//std::cout<<--c<<std::endl;
						//c--;
						//j=temp;
						j--;
						//						std::cout<<"targets j-1 "<<(*(j--)).x<<","<<(*(j)).y
						//								<<"-- j "<<(*(j++)).x<<","<<(*(j)).y
						//								<<"-- j+1 "<<(*(j++)).x<<","<<(*(j)).y<<std::endl;
						//						copy(targets.begin(), targets.end(), v_targets.begin());
						//
						//						dlog.addText ( Logger::TEAM,__FILE__": targets %d (%f,%f) -- %d (%f,%f) -- %d (%f,%f)",
						//								i-1,v_targets[i-1].x,v_targets[i-1].y,
						//								i,v_targets[i].x,v_targets[i].y,
						//								i+1,v_targets[i+1].x,v_targets[i+1].y);//						j--;
						//e--;
					}
					j++;
				}

				//if(targets.size()>0 && !isPosition(*(targets.begin()),wm)) targets.erase(targets.begin());


				break;
			}

			default:;
			}
		}


		if(targets.size()==0 )
		{
			dlog.addText ( Logger::TEAM,__FILE__": No targets. Relax predicates");

			//delete one predicate which erased most targets
			int max=0;
			int max_i=-1;
			for(int i=0;i<o_activePredicates.size();i++)
			{
				if(d[i]>max)
				{
					max=d[i];
					max_i=i;
				}
			}

			//delete max_i from the predicates and try again
			if(max_i!=-1 && o_activePredicates.size()>=1)
			{
				o_activePredicates.erase(o_activePredicates.begin()+max_i);

				return generateTargets(wm,targets,offence);
			}
			else
				return;


		}
		//
		//
	}
//	else 	//we do not have the ball
//	{
//		if(d_activePredicates.size()==0) return;
//		for(uint i=0;i<d_activePredicates.size();i++)
//		{
//			switch(d_activePredicates[i])
//			{
//			case blocked:
//			{
//				if(isBlocked(wm.self().pos(),wm))
//					break;
//				std::list<rcsc::Vector2D>::iterator j=targets.begin();
//				int c=targets.size();
//				//std::cout<<"blocked: "<<c<<std::endl;
//				while(true)
//				{
//					j++;
//					if(j==targets.end()) break;
//					//check locations that keep opp blocked
//					if(!isBlocked(*j,wm))
//					{
//						//std::list<rcsc::Vector2D>::iterator temp=j;
//
//						targets.erase(j);
//						//std::cout<<--c<<std::endl;
//						c--;
//						j--;
//						//j=temp;
//					}
//				}
//				break;
//			}
//			case marked:
//			{
//				std::list<rcsc::Vector2D>::iterator j=targets.begin();
//				int c=targets.size();
//				//std::cout<<"marked: "<<c<<std::endl;
//				while(true)
//				{
//					j++;
//					if(j==targets.end()) break;
//					//check locations that keep opp marked
//					if(!isMarked(*j,wm))
//					{
//						//std::list<rcsc::Vector2D>::iterator temp=j;
//
//						targets.erase(j);
//						//std::cout<<--c<<std::endl;
//						c--;
//						j--;
//						//j=temp;
//					}
//				}
//				break;
//			}
//			case supported:
//			{
//				std::list<rcsc::Vector2D>::iterator j=targets.begin();
//				int c=targets.size();
//				std::cout<<"supported: "<<c<<std::endl;
//				while(true)
//				{
//					j++;
//					if(j==targets.end()) break;
//					//check locations that keep support
//					if(!isSupported(*j,wm))
//					{
//						//std::list<rcsc::Vector2D>::iterator temp=j;
//
//						targets.erase(j);
//						//std::cout<<--c<<std::endl;
//						c--;
//						j--;
//						//j=temp;
//					}
//				}
//				break;
//			}
//			case tacklable:
//			{
//				std::list<rcsc::Vector2D>::iterator j=targets.begin();
//				int c=targets.size();
//				std::cout<<"tacklable: "<<c<<std::endl;
//				while(true)
//				{
//					j++;
//					if(j==targets.end()) break;
//					//check locations that keep you tacklable
//					if(!isTacklable(*j,wm))
//					{
//						//std::list<rcsc::Vector2D>::iterator temp=j;
//
//						targets.erase(j);
//						//std::cout<<--c<<std::endl;
//						c--;
//						j--;
//						//j=temp;
//					}
//				}
//				break;
//			}
//
//			default:;
//
//			}

//		}
//	}



	//if any targets remain in the vector, choose from them and act
	//return targets;
}

/*-------------------------------------------------------------------*/

bool PredicateComplex::isPosession(Vector2D t,const rcsc::WorldModel& wm )
{
    const int self_min = wm.interceptTable()->selfReachCycle();
    //const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();
    std::cout<<"self_min: "<<self_min<<" - opp_min: "<<opp_min<<std::endl;
    if (wm.self().isKickable() || self_min<opp_min) //|| !wm.existKickableOpponent()
	    return true;
    return false;

}

/*-------------------------------------------------------------------*/

bool PredicateComplex::isOpen(Vector2D t,const rcsc::WorldModel& wm )
{
	int OPENNESS_THRESHOLD=5;

	Vector2D me_targ=t-wm.self().pos();

	if(me_targ.norm()>=2)
	{
		double p2x=t.x;
		double p2y=t.y;

		double p1x=wm.self().pos().x;
		double p1y=wm.self().pos().y;

		double p3x=p2x-p1x;
		double p3y=p2y-p1y;


		me_targ.setLength(me_targ.length()+2);

		p3x=me_targ.x;
		p3y=me_targ.y;

		//scaling factor
		double f=0.33*t.dist(wm.self().pos());


		//std::cout<<"fromto: "<<p3x<<", "<<p3y<<std::endl;
		double p31x=-1*p3y;
		double p31y=p3x;

		p31x/=sqrt(p31x*p31x+p31y*p31y)/f;
		p31y/=sqrt(p31x*p31x+p31y*p31y)/f;
		//std::cout<<"first point: "<<p31x<<", "<<p31y<<std::endl;

		double p4x=p1x+p3x+p31x;
		double p4y=p1y+p3y+p31y;
		//std::cout<<"first point: "<<p4x<<", "<<p4y<<std::endl;
		double p32x=p3y;
		double p32y=-1*p3x;


		p32x/=sqrt(p32x*p32x+p32y*p32y)/f;
		p32y/=sqrt(p32x*p32x+p32y*p32y)/f;
		//std::cout<<"second point: "<<p32x<<", "<<p32y<<std::endl;

		double p5x=p1x+p3x+p32x;
		double p5y=p1y+p3y+p32y;
		//std::cout<<"second point: "<<p5x<<", "<<p5y<<std::endl;
		int c_o=wm.countOpponentsIn<Triangle2D>(Triangle2D(/*Vector2D(p1x,p1y)*/wm.self().pos(),
				Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);


		//int c_m=wm.countTeammatesIn<Triangle2D>(Triangle2D(/*Vector2D(p1x,p1y)*/wm.self().pos(),
			//			Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);

		int c_m=wm.countTeammatesIn<Circle2D>(Circle2D(t,10),
								//Vector2D(p4x,p4y),Vector2D(p5x,p5y))
								100,true);


		if (c_o==0 && c_m <= 1)// && (t-Strategy::i().getPosition( wm.self().unum() )).norm<28) //if the path is clear
		{
			return true;
		}
		if(c_o<=1 && c_m<=2 && t.dist(wm.self().pos())>8)
		{
			return true;
		}
	}

	return false;



//	if(wm.getDistOpponentNearestTo(t,1)<OPENNESS_THRESHOLD)
//		return false;
//	else if(wm.countOpponentsIn<Circle2D>(Circle2D(t,10),1,true)>1)
//		return false;
//	return true;
}

/*-------------------------------------------------------------------*/

bool PredicateComplex::isPosition(Vector2D t,const rcsc::WorldModel& wm )
{
	//Strategy::i().get .getPosition() Strategy::i().roleNumber(wm.self().unum());
	if(!t.isValid()) //check if the way is open to advance into better positions
	{
		Vector2D goal_c= ServerParam::instance().theirTeamGoalPos();

		Vector2D me_goal=goal_c-wm.self().pos();
		Vector2D dist1=me_goal.rotatedVector(90);
		dist1.normalize();
		dist1.scale(3);
		Vector2D dist2=me_goal.rotatedVector(-90);
		dist2.normalize();
		dist2.scale(3);
		Vector2D dist3(0,0.5*ServerParam::i().DEFAULT_GOAL_WIDTH);
		Vector2D dist4(0,-0.5*ServerParam::i().DEFAULT_GOAL_WIDTH);


		//the box of immediate advancing toward the goal
		Vector2D p1=wm.self().pos()+dist1;
		Vector2D p2=wm.self().pos()+dist2;
		Vector2D p3=goal_c+dist3;
		Vector2D p4=goal_c+dist4;

		std::vector<Vector2D> ps;
		ps.push_back(p1);
		ps.push_back(p2);
		ps.push_back(p3);
		ps.push_back(p4);

		return (wm.countOpponentsIn(Polygon2D(ps),100,false)<4);//Rect2D(p1,p4)
	}
	Vector2D me_goal=ServerParam::instance().theirTeamGoalPos()-wm.self().pos();
	Vector2D t_goal=ServerParam::instance().theirTeamGoalPos()-t;


    return (me_goal.norm()  > t_goal.norm()  &&  t.x<=(wm.self().unum()<=6?((wm.self().unum()-1)/(10))*ServerParam::i().DEFAULT_PITCH_LENGTH/2-17:ServerParam::i().DEFAULT_PITCH_LENGTH/2));
        //new position closer to goal than old position
}




bool PredicateComplex::isContact(Vector2D t,const rcsc::WorldModel& wm )
{
	bool goalie_possible=false;
	int mate_possible=-1;
	int mate_possible_behind=-1;

	if (wm.self().isKickable())
	{
		//search for any contact

		//start with clear contact with goal
		double p1x=t.x;
		double p1y=t.y;
		double p2x=ServerParam::i().theirTeamGoalPos().x;
		double p2y=ServerParam::i().theirTeamGoalPos().y;
		double p3x=p2x-p1x;
		double p3y=p2y-p1y;
		double p31x=-1*p3y;
		double p31y=p3x;

		//scaling factor
		double f=0.33*t.dist(ServerParam::i().theirTeamGoalPos());

		p31x/=sqrt(p31x*p31x+p31y*p31y)/f;
		p31y/=sqrt(p31x*p31x+p31y*p31y)/f;
		double p4x=p2x+p31x;
		double p4y=p2y+p31y;
		double p32x=p3y;
		double p32y=-1*p3x;


		p32x/=sqrt(p32x*p32x+p32y*p32y)/f;
		p32y/=sqrt(p32x*p32x+p32y*p32y)/f;
		double p5x=p2x+p32x;
		double p5y=p2y+p32y;
		int c=wm.countOpponentsIn<Triangle2D>(Triangle2D(/*Vector2D(p1x,p1y)*/wm.ball().pos(),
				Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);

		if (c==0) //if the path is clear
		{
			if(t.x==wm.self().pos().x && t.y==wm.self().pos().y)
			{
				pass_mate=0;
				dlog.addText ( Logger::TEAM,__FILE__": pass_mate %d (%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)",pass_mate,wm.ball().pos().x,wm.ball().pos().y,p4x,p4y,p5x,p5y);
			}
			return true;

		}

		//continue with teammates
		std::vector< PlayerObject*>::const_iterator teammate=wm.teammatesFromSelf().begin();

		while (teammate!=wm.teammatesFromSelf().end())
		{
			Vector2D me_mate=(*teammate)->pos()-t;

			if(me_mate.norm()>=3)
			{
				double p1x=t.x;
				double p1y=t.y;
				//std::cout<<"target: "<<p1x<<", "<<p1y<<std::endl;
				double p2x=(*teammate)->pos().x;
				double p2y=(*teammate)->pos().y;
				//std::cout<<"mate: "<<p2x<<", "<<p2y<<std::endl;
				double p3x=p2x-p1x;
				double p3y=p2y-p1y;


				me_mate.setLength(me_mate.length()+5);

				p3x=me_mate.x;
				p3y=me_mate.y;

				//scaling factor
				double f=0.6*t.dist((*teammate)->pos());


				//std::cout<<"fromto: "<<p3x<<", "<<p3y<<std::endl;
				double p31x=-1*p3y;
				double p31y=p3x;

				p31x/=sqrt(p31x*p31x+p31y*p31y)/f;
				p31y/=sqrt(p31x*p31x+p31y*p31y)/f;
				//std::cout<<"first point: "<<p31x<<", "<<p31y<<std::endl;

				double p4x=p1x+p3x+p31x;
				double p4y=p1y+p3y+p31y;
				//std::cout<<"first point: "<<p4x<<", "<<p4y<<std::endl;
				double p32x=p3y;
				double p32y=-1*p3x;


				p32x/=sqrt(p32x*p32x+p32y*p32y)/f;
				p32y/=sqrt(p32x*p32x+p32y*p32y)/f;
				//std::cout<<"second point: "<<p32x<<", "<<p32y<<std::endl;

				double p5x=p1x+p3x+p32x;
				double p5y=p1y+p3y+p32y;
				//std::cout<<"second point: "<<p5x<<", "<<p5y<<std::endl;
				int c=wm.countOpponentsIn<Triangle2D>(Triangle2D(/*Vector2D(p1x,p1y)*/wm.ball().pos(),
						Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);

				if (c==0) //if the path is clear
				{

					const int goalie_unum = ( wm.teammateGoalieUnum() != Unum_Unknown ? wm.teammateGoalieUnum() : Strategy::i().goalieUnum() );


					if((*teammate)->unum()==goalie_unum)	//prefer not to pass to goalie if other options are available
					{
						goalie_possible=true;
					}
					else
					{
						if((*teammate)->pos().x < t.x /*&& rand()%100<90*/) //100% of the time if mate is behind
						{
							if(mate_possible_behind==-1)	//push forward
							{
								mate_possible_behind=(*teammate)->unum();	//save it for later
							}
						}
						else
						{
							if(rand()%100<40)	//randomize in pass selection
							{
								mate_possible=(*teammate)->unum();	//save it for later
							}
							else
							{
								if(t.x==wm.self().pos().x && t.y==wm.self().pos().y)
								{
									pass_mate=(*teammate)->unum();
									dlog.addText ( Logger::TEAM,__FILE__": pass_mate %d (%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)",pass_mate,wm.ball().pos().x,wm.ball().pos().y,p4x,p4y,p5x,p5y);
								}
								return true;
							}

						}
					}
				}
			}
			//else check next teammate
			//std::cout<<"triangle not clear!"<<std::endl;
			teammate++;
		}

		//reached here without choosing
		if(mate_possible!=-1)
		{
			if(t.x==wm.self().pos().x && t.y==wm.self().pos().y)
			{
				pass_mate=mate_possible;
				dlog.addText ( Logger::TEAM,__FILE__": pass_mate (possible) %d (%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)",pass_mate,wm.ball().pos().x,wm.ball().pos().y,p4x,p4y,p5x,p5y);
			}
			return true;
		}
		if(mate_possible_behind!=-1)// && rand()%100>60)
		{
			if(t.x==wm.self().pos().x && t.y==wm.self().pos().y)
			{
				pass_mate=mate_possible_behind;
				dlog.addText ( Logger::TEAM,__FILE__": pass_mate (behind) %d (%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)",pass_mate,wm.ball().pos().x,wm.ball().pos().y,p4x,p4y,p5x,p5y);
			}
			return true;
		}
		if(goalie_possible && rand()%100>90)
		{
			if(t.x==wm.self().pos().x && t.y==wm.self().pos().y)
			{
				const int goalie_unum = ( wm.teammateGoalieUnum() != Unum_Unknown ? wm.teammateGoalieUnum() : Strategy::i().goalieUnum() );
				pass_mate=goalie_unum;
				dlog.addText ( Logger::TEAM,__FILE__": pass_mate (goalie) %d (%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)",pass_mate,wm.ball().pos().x,wm.ball().pos().y,p4x,p4y,p5x,p5y);
			}
			return true;
		}
		return false;
	}
	else
	{
		//search for contact with  ball holder only
		if(wm.allTeammates().size()==1) return false;

		const PlayerObject* teammate=wm.getTeammateNearestToBall(1,true);

		if (teammate)
		{
			double p2x=t.x;
			double p2y=t.y;

			double p1x=teammate->pos().x;
			double p1y=teammate->pos().y;

			double p3x=p2x-p1x;
			double p3y=p2y-p1y;

			double p31x=-1*p3y;
			double p31y=p3x;

			p31x/=sqrt(p31x*p31x+p31y*p31y)/5;
			p31y/=sqrt(p31x*p31x+p31y*p31y)/5;


			double p4x=p2x+p31x;
			double p4y=p2y+p31y;

			double p32x=p3y;
			double p32y=-1*p3x;

			p32x/=sqrt(p32x*p32x+p32y*p32y)/5;
			p32y/=sqrt(p32x*p32x+p32y*p32y)/5;

			double p5x=p2x+p32x;
			double p5y=p2y+p32y;

			int c=wm.countOpponentsIn<Triangle2D>(Triangle2D(Vector2D(p1x,p1y),Vector2D(p4x,p4y),Vector2D(p5x,p5y)),100,true);

			if (c==0) //if the path is clear
					return true;
		}
		return false;
	}
}


/*-------------------------------------------------------------------*/
//check if the opponent way to goal is properly blocked
bool PredicateComplex::isBlocked(Vector2D t,const rcsc::WorldModel& wm )
{
	//dlog.addText ( Logger::TEAM,__FILE__": isBlocked");
	bool block_possible=false;
	const PlayerObject * m;

	bool shooter=false;
	if(wm.getOpponentNearestToBall(100,true))shooter=wm.getOpponentNearestToBall(100,true)->pos().dist(ServerParam::i().ourTeamGoalPos())<15;

	if(
			wm.getOpponentNearestToBall(100,true)
			&& ( //consider the keeper as a blocker if too close to goal
					(shooter && (m=wm.getTeammateNearestToBall(100,true)))
					||
					(m=wm.getTeammateNearestToBall(100,false))
			)
			&& (m &&  (wm.getOpponentNearestToBall(100,true))->distFromBall() >= m->distFromBall())
	)
	{
		dlog.addText ( Logger::TEAM,__FILE__": opp %d -- blocker mate %d",wm.getOpponentNearestToBall(100,true)->unum(),m->unum());
		return true;
	}

	//is there a dangerous striker?
	if(wm.getOpponentNearestToBall(100,true))// && (wm.getOpponentNearestToBall(100,true))->distFromBall()<10)
	{
		//get opponent position
		Vector2D opp_pos=(wm.getOpponentNearestToBall(100,true))->pos();
		Vector2D opp_dir(0,0);
		opp_dir.setPolar(1,(wm.getOpponentNearestToBall(100,true))->body());
		//int opp_unum=(wm.getOpponentNearestToBall(100,true))->unum();

		//		dlog.addText ( Logger::TEAM,
		//			                   __FILE__": isBlocked--opp %d (%f,%f) - opp_dir %f",opp_unum,opp_pos.x,opp_pos.y,opp_dir.dir().degree());

		//
		//	//check if there is a blocking player
		//	Vector2D goal_c= ServerParam::instance().ourTeamGoalPos();
		//
		//	Vector2D dist=goal_c;
		//	dist=opp_pos-dist;
		//	Vector2D dist1=dist.rotatedVector(90);
		//	dist1.normalize();
		//	dist1.scale(5);
		//	Vector2D dist2=dist.rotatedVector(-90);
		//	dist2.normalize();
		//	dist2.scale(5);
		//	Vector2D p1=dist+dist1;
		//	Vector2D p2=dist+dist2;
		//	Vector2D p3=goal_c+dist1;
		//	Vector2D p4=goal_c+dist2;

		//check if there is a blocking player
		Vector2D goal_c= ServerParam::instance().ourTeamGoalPos();

		Vector2D opp_goal=goal_c-opp_pos;
		Vector2D dist1=opp_goal.rotatedVector(90);
		dist1.normalize();
		dist1.scale(3);
		Vector2D dist2=opp_goal.rotatedVector(-90);
		dist2.normalize();
		dist2.scale(3);
		Vector2D dist3(0,0.5*ServerParam::i().DEFAULT_GOAL_WIDTH);
		Vector2D dist4(0,-0.5*ServerParam::i().DEFAULT_GOAL_WIDTH);


		//the box of blocking for the ball holding opponent
		Vector2D p1=opp_pos+dist1;
		Vector2D p2=opp_pos+dist2;
		Vector2D p3=goal_c+dist3;
		Vector2D p4=goal_c+dist4;


		std::vector<Vector2D> ps(4);
		ps.push_back(p1);
		ps.push_back(p2);
		ps.push_back(p3);
		ps.push_back(p4);


		//		dlog.addText ( Logger::TEAM,__FILE__": isBlocked--p1 (%f,%f) - p2 (%f,%f) - p3 (%f,%f) - p4 (%f,%f) -- exists? %d",p1.x,p1.y,p2.x,p2.y,p3.x,p3.y,p4.x,p4.y,(wm.existTeammateIn(Rect2D(p1,p4),100,false))?1:0);

		//path blocked if somebody is already blocking the path, or if current movement would lead to blocking
		const int goalie_unum = ( wm.teammateGoalieUnum() != Unum_Unknown ? wm.teammateGoalieUnum() : Strategy::i().goalieUnum() );

		if(wm.existTeammateIn(Polygon2D(ps),100,false)
				|| (  shooter  && wm.teammate(goalie_unum) &&
						(Polygon2D(ps)).contains(wm.teammate(goalie_unum)->pos()) )  )//Rect2D(p1,p4)
		{
			dlog.addText ( Logger::TEAM,__FILE__": in blocking area (%.1f,%.1f),(%.1f,%.1f),(%.1f,%.1f),(%.1f,%.1f)",p1.x,p1.y,p2.x,p2.y,p3.x,p3.y,p4.x,p4.y);
			int i=(((wm.getOpponentNearestToBall(100,true))->pos().dist(ServerParam::i().ourTeamGoalPos())<15)?1:2);
			for(;i<=11;i++)
			{
//				if(i==wm.self().unum())
//					continue;
				if(wm.teammate(i) && (Polygon2D(ps)).contains(wm.teammate(i)->pos()))
					break;
			}

			if(i<=11 && Marked(0))
			{
				dlog.addText ( Logger::TEAM,__FILE__": blocking -- mate %d (or %d)",i,Marker(0));
				return true;
			}
			else
			{
				dlog.addText ( Logger::TEAM,__FILE__": blocking -- claimed but not proven");

				return false;
			}



		}

		//no teammates in blocking box

		//consider blocking
		//t is new position, direction extracted from current position
		//will continuing in the same direction be enough to block opp?
		Vector2D my_dirv=t-wm.self().pos();

		//does t have useful information?
		if((t.x==wm.self().pos().x && t.y==wm.self().pos().y) || (t.x==0 && t.y==0))
		{
			//no useful position information, use the best possible block
			//my_dirv.setLength(1);
			Vector2D me_goal=goal_c-wm.self().pos();
			Vector2D me_opp=opp_pos-wm.self().pos();

			double diff=fabs(me_goal.dir().degree()-me_opp.dir().degree());

			double angle=0;
			if(diff>180)
			{
				diff=360-diff;
				//the angles have different signs
				if (me_goal.dir().degree()>me_opp.dir().degree()) 	//positive
				{
					if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
						angle=me_goal.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()-diff/2;
				}
				else
				{
					if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
						angle=me_goal.dir().degree()-diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
			}
			else
			{
				if(me_goal.dir().degree()*me_opp.dir().degree() >0) 	//of same sign
				{
					if(me_goal.dir().degree()<me_opp.dir().degree()) //smaller
						angle=me_goal.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
				else	//different signs
				{
					if(me_goal.dir().degree()<me_opp.dir().degree()) //negative
						angle=me_goal.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
			}

			my_dirv.setPolar(1,AngleDeg(angle));

			//			if(fabs(dist1.dir().degree()-me_goal.dir().degree())<fabs(dist2.dir().degree()-me_goal.dir().degree()))
			//				my_dirv.setPolar(1,dist1.dir())	;
			//			else
			//				my_dirv.setPolar(1,dist2.dir())	;
		}

		//dirv contains the proposed blocking action

		my_dirv.normalize();



		//dlog.addText ( Logger::TEAM,
		//                   __FILE__": blocking for %f (t is (%f,%f) and pos is (%f,%f))",my_dirv.dir().degree(),t.x,t.y,wm.self().pos().x,wm.self().pos().y);

		Vector2D distv=wm.self().pos() - opp_pos;
		//double dist=distv.norm();

		//the opponent direction is not toward the goal, assume it is
		if(fabs(opp_goal.dir().degree()-opp_dir.dir().degree())>45)
			opp_dir.setPolar(1,opp_goal.dir().degree());

		Vector2D velv=my_dirv-opp_dir;
		//dobule vel=velv.norm();

		//		dlog.addText ( Logger::TEAM,
		//			                   __FILE__": distv (%f,%f) - velv (%f,%f)",distv.x,distv.y,velv.x,velv.y);

		//		if(fabs(distv.dir().degree()-velv.dir().degree()) >= 0 && fabs(distv.dir().degree()-velv.dir().degree()) <90)
		//			block_possible=true;

		//would the proposed action cause a proper block after some time?
		if((opp_pos.y-wm.self().pos().y)*cos(opp_dir.dir().radian()) -  (opp_pos.x-wm.self().pos().x)*sin(opp_dir.dir().radian())
				< (opp_pos.y-wm.self().pos().y)*cos(wm.self().body().radian()) -  (opp_pos.x-wm.self().pos().x)*sin(wm.self().body().radian()))
		{
			//dlog.addText ( Logger::TEAM,__FILE__": blocking possible");

			block_possible=true;
		}

		//
		//	    dlog.addText ( Logger::TEAM,
		//	                   __FILE__": blocking for direction %f is %d",my_dirv.dir().degree(), (block_possible)?1:0);


		///////////// return block_possible;
		//Rect2D(p1,p4)//wm.existTeammateIn(Polygon2D(ps),100,false) ||
	}
	//else
		return false;
}


/*-------------------------------------------------------------------*/
//check if the opponent way to goal is properly blocked
bool PredicateComplex::isMarked(Vector2D t,const rcsc::WorldModel& wm )
{

	//get offensive opponent positions
//	rcsc::XCoordinateForwardPlayerPredicate x(0);
//	AbstractPlayerCont opps=wm.getPlayerCont((PlayerPredicate*)(&x));
	if(t.isValid())
	{
		//check if at position t the agent is participating in marking
		double dist=0;
		wm.getTeammateNearestTo(wm.getOpponentNearestToSelf(100,false),100,&dist);
		if(dist<wm.getDistOpponentNearestToSelf(100,false))
		{	//I'm closer to that opponent
			//is the position of the agent proper to mark the opponent?
			Vector2D opp_pos=(wm.getOpponentNearestToSelf(100,false))->pos();

			Vector2D me_ball=wm.ball().pos()-wm.self().pos();
			Vector2D me_opp=opp_pos-wm.self().pos();

			double diff=fabs(me_ball.dir().degree()-me_opp.dir().degree());

			double angle=0;
			if(diff>180)
			{
				diff=360-diff;
				//the angles have different signs
				if (me_ball.dir().degree()>me_opp.dir().degree()) 	//positive
				{
					if(fabs(me_ball.dir().degree())<fabs(me_opp.dir().degree())) //smaller
						angle=me_ball.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()-diff/2;
				}
				else
				{
					if(fabs(me_ball.dir().degree())<fabs(me_opp.dir().degree())) //smaller
						angle=me_ball.dir().degree()-diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
			}
			else
			{
				if(me_ball.dir().degree()*me_opp.dir().degree() >0) 	//of same sign
				{
					if(me_ball.dir().degree()<me_opp.dir().degree()) //smaller
						angle=me_ball.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
				else	//different signs
				{
					if(me_ball.dir().degree()<me_opp.dir().degree()) //negative
						angle=me_ball.dir().degree()+diff/2;
					else
						angle=me_opp.dir().degree()+diff/2;
				}
			}

			//diff is the angle separating the opponent from the ball
			return(abs(diff)>160);	//close to the blocking position


		}
		else return true;
	}
	else	//check the general condition
	{
		AbstractPlayerCont opps=wm.getPlayerCont((PlayerPredicate*)(new rcsc::XCoordinateForwardPlayerPredicate(0)));

		std::vector<const rcsc::AbstractPlayerObject*>::iterator opp=opps.begin();

		while(opp!=opps.end())
		{
			std::cout<<(*opp)->pos()<<std::endl;
			if(!wm.existTeammateIn(Circle2D((*opp)->pos(),10),100,true))
				return false;
			opp++;
		}
		return true;

	}
}



/*-------------------------------------------------------------------*/
//check if the opponent way to goal is properly blocked
bool PredicateComplex::isSupported(Vector2D t,const rcsc::WorldModel& wm )
{
	return false;
}




/*-------------------------------------------------------------------*/
//check if the opponent way to goal is properly blocked
bool PredicateComplex::isTacklable(Vector2D t,const rcsc::WorldModel& wm )
{
	return false;
}


